@@ -4,7 +4,6 @@ I/O for VTU.
 <https://vtk.org/wp-content/uploads/2015/04/file-formats.pdf>
 """
 import base64
-import logging
 import re
 import sys
 import zlib
@@ -12,9 +11,9 @@ import zlib
 import numpy as np
 
 from ..__about__ import __version__
-from .._common import raw_from_cell_data
-from .._exceptions import ReadError
-from .._helpers import register
+from .._common import raw_from_cell_data, warn
+from .._exceptions import CorruptionError, ReadError
+from .._helpers import register_format
 from .._mesh import CellBlock, Mesh
 from .._vtk_common import meshio_to_vtk_order, meshio_to_vtk_type, vtk_cells_from_data
 
@@ -191,7 +190,9 @@ def _parse_raw_binary(filename):
         raw = f.read()
 
     try:
-        i_start = re.search(re.compile(b'<AppendedData[^>]+(?:">)'), raw).end()
+        res = re.search(re.compile(b'<AppendedData[^>]+(?:">)'), raw)
+        assert res is not None
+        i_start = res.end()
         i_stop = raw.find(b"</AppendedData>")
     except Exception:
         raise ReadError()
@@ -209,10 +210,30 @@ def _parse_raw_binary(filename):
         )
 
     appended_data_tag = root.find("AppendedData")
+    assert appended_data_tag is not None
     appended_data_tag.set("encoding", "base64")
 
     compressor = root.get("compressor")
-    if compressor is not None:
+    if compressor is None:
+        arrays = ""
+        i = 0
+        while i < len(data):
+            # The following find() runs into issues if offset is padded with spaces, see
+            # <https://github.com/nschloe/meshio/issues/1135>. It works in ParaView.
+            # Unfortunately, Python's built-in XML tree can't handle regexes, see
+            # <https://stackoverflow.com/a/38810731/353337>.
+            da_tag = root.find(f".//DataArray[@offset='{i}']")
+            if da_tag is None:
+                raise RuntimeError(f"Could not find .//DataArray[@offset='{i}']")
+            da_tag.set("offset", str(len(arrays)))
+
+            block_size = int(np.frombuffer(data[i : i + dtype.itemsize], dtype)[0])
+            arrays += base64.b64encode(
+                data[i : i + block_size + dtype.itemsize]
+            ).decode()
+            i += block_size + dtype.itemsize
+
+    else:
         c = {"vtkLZMADataCompressor": lzma, "vtkZLibDataCompressor": zlib}[compressor]
         root.attrib.pop("compressor")
 
@@ -220,8 +241,9 @@ def _parse_raw_binary(filename):
         arrays = ""
         i = 0
         while i < len(data):
-            da_tag = root.find(".//DataArray[@offset='%d']" % i)
-            da_tag.set("offset", "%d" % len(arrays))
+            da_tag = root.find(f".//DataArray[@offset='{i}']")
+            assert da_tag is not None
+            da_tag.set("offset", str(len(arrays)))
 
             num_blocks = int(np.frombuffer(data[i : i + dtype.itemsize], dtype)[0])
             num_header_items = 3 + num_blocks
@@ -244,19 +266,6 @@ def _parse_raw_binary(filename):
 
             i += j + num_header_bytes
 
-    else:
-        arrays = ""
-        i = 0
-        while i < len(data):
-            da_tag = root.find(".//DataArray[@offset='%d']" % i)
-            da_tag.set("offset", "%d" % len(arrays))
-
-            block_size = int(np.frombuffer(data[i : i + dtype.itemsize], dtype)[0])
-            arrays += base64.b64encode(
-                data[i : i + block_size + dtype.itemsize]
-            ).decode()
-            i += block_size + dtype.itemsize
-
     appended_data_tag.text = "_" + arrays
     return root
 
@@ -293,9 +302,10 @@ class VtuReader:
             root = _parse_raw_binary(str(filename))
 
         if root.tag != "VTKFile":
-            raise ReadError()
+            raise ReadError(f"Expected tag 'VTKFile', found {root.tag}")
         if root.attrib["type"] != "UnstructuredGrid":
-            raise ReadError()
+            tpe = root.attrib["type"]
+            raise ReadError(f"Expected type UnstructuredGrid, found {tpe}")
 
         if "version" in root.attrib:
             version = root.attrib["version"]
@@ -390,7 +400,10 @@ class VtuReader:
                     for c in child:
                         if c.tag != "DataArray":
                             raise ReadError()
-                        piece_point_data[c.attrib["Name"]] = self.read_data(c)
+                        try:
+                            piece_point_data[c.attrib["Name"]] = self.read_data(c)
+                        except CorruptionError as e:
+                            warn(e.args[0] + " Skipping.")
 
                     point_data.append(piece_point_data)
 
@@ -433,23 +446,22 @@ class VtuReader:
     def read_uncompressed_binary(self, data, dtype):
         byte_string = base64.b64decode(data)
 
+        # the first item is the total_num_bytes, given in header_dtype
         header_dtype = vtu_to_numpy_type[self.header_type]
         if self.byte_order is not None:
             header_dtype = header_dtype.newbyteorder(
                 "<" if self.byte_order == "LittleEndian" else ">"
             )
-        num_bytes_per_item = np.dtype(header_dtype).itemsize
-        total_num_bytes = int(
-            np.frombuffer(byte_string[:num_bytes_per_item], header_dtype)[0]
-        )
+        num_header_bytes = np.dtype(header_dtype).itemsize
+        total_num_bytes = np.frombuffer(byte_string[:num_header_bytes], header_dtype)[0]
 
         # Check if block size was decoded separately
         # (so decoding stopped after block size due to padding)
-        if len(byte_string) == num_bytes_per_item:
+        if len(byte_string) == num_header_bytes:
             header_len = len(base64.b64encode(byte_string))
             byte_string = base64.b64decode(data[header_len:])
         else:
-            byte_string = byte_string[num_bytes_per_item:]
+            byte_string = byte_string[num_header_bytes:]
 
         # Read the block data; multiple blocks possible here?
         if self.byte_order is not None:
@@ -541,12 +553,22 @@ class VtuReader:
                 if self.compression is None
                 else self.read_compressed_binary
             )
+            assert self.appended_data is not None
             data = reader(self.appended_data[offset:], dtype)
         else:
             raise ReadError(f"Unknown data format '{fmt}'.")
 
         if "NumberOfComponents" in c.attrib:
-            data = data.reshape(-1, int(c.attrib["NumberOfComponents"]))
+            nc = int(c.attrib["NumberOfComponents"])
+            try:
+                data = data.reshape(-1, nc)
+            except ValueError:
+                name = c.attrib["Name"]
+                raise CorruptionError(
+                    "VTU file corrupt. "
+                    + f"The size of the data array '{name}' is {data.size} "
+                    + f"which doesn't fit the number of components {nc}."
+                )
         return data
 
 
@@ -592,16 +614,16 @@ def write(filename, mesh, binary=True, compression="zlib", header_type=None):
                 )
 
     if not binary:
-        logging.warning("VTU ASCII files are only meant for debugging.")
+        warn("VTU ASCII files are only meant for debugging.")
 
     if mesh.points.shape[1] == 2:
-        logging.warning(
+        warn(
             "VTU requires 3D points, but 2D points given. "
             "Appending 0 third component."
         )
-        mesh.points = np.column_stack(
-            [mesh.points[:, 0], mesh.points[:, 1], np.zeros(mesh.points.shape[0])]
-        )
+        points = np.column_stack([mesh.points, np.zeros_like(mesh.points[:, 0])])
+    else:
+        points = mesh.points
 
     vtk_file = ET.Element(
         "VTKFile",
@@ -611,13 +633,15 @@ def write(filename, mesh, binary=True, compression="zlib", header_type=None):
         # a bit.
         byte_order=("LittleEndian" if sys.byteorder == "little" else "BigEndian"),
     )
-    header_type = (
-        "UInt32" if header_type is None else vtk_file.set("header_type", header_type)
-    )
+
+    if header_type is None:
+        header_type = "UInt32"
+    else:
+        vtk_file.set("header_type", header_type)
     assert header_type is not None
 
     if binary and compression:
-        # TODO lz4, lzma <https://vtk.org/doc/nightly/html/classvtkDataCompressor.html>
+        # TODO lz4 <https://vtk.org/doc/nightly/html/classvtkDataCompressor.html>
         compressions = {
             "lzma": "vtkLZMADataCompressor",
             "zlib": "vtkZLibDataCompressor",
@@ -628,8 +652,10 @@ def write(filename, mesh, binary=True, compression="zlib", header_type=None):
     # swap the data to match the system byteorder
     # Don't use byteswap to make sure that the dtype is changed; see
     # <https://github.com/numpy/numpy/issues/10372>.
-    points = mesh.points.astype(mesh.points.dtype.newbyteorder("="), copy=False)
-    for k, (cell_type, data) in enumerate(mesh.cells):
+    points = points.astype(points.dtype.newbyteorder("="), copy=False)
+    for k, cell_block in enumerate(mesh.cells):
+        cell_type = cell_block.type
+        data = cell_block.data
         # Treatment of polyhedra is different from other types
         if is_polyhedron_grid:
             new_cell_info = []
@@ -661,70 +687,65 @@ def write(filename, mesh, binary=True, compression="zlib", header_type=None):
         da = ET.SubElement(parent, "DataArray", type=vtu_type, Name=name)
         if len(data.shape) == 2:
             da.set("NumberOfComponents", f"{data.shape[1]}")
+
+        def text_writer_compressed(f):
+            max_block_size = 32768
+            data_bytes = data.tobytes()
+
+            # round up
+            num_blocks = -int(-len(data_bytes) // max_block_size)
+            last_block_size = len(data_bytes) - (num_blocks - 1) * max_block_size
+
+            # It's too bad that we have to keep all blocks in memory. This is
+            # necessary because the header, written first, needs to know the
+            # lengths of all blocks. Also, the blocks are encoded _after_ having
+            # been concatenated.
+            c = {"lzma": lzma, "zlib": zlib}[compression]
+            compressed_blocks = [
+                # This compress is the slowest part of the writer
+                c.compress(block)
+                for block in _chunk_it(data_bytes, max_block_size)
+            ]
+
+            # collect header
+            header = np.array(
+                [num_blocks, max_block_size, last_block_size]
+                + [len(b) for b in compressed_blocks],
+                dtype=vtu_to_numpy_type[header_type],
+            )
+            f.write(base64.b64encode(header.tobytes()).decode())
+            f.write(base64.b64encode(b"".join(compressed_blocks)).decode())
+
+        def text_writer_uncompressed(f):
+            data_bytes = data.tobytes()
+            # collect header
+            header = np.array(len(data_bytes), dtype=vtu_to_numpy_type[header_type])
+            f.write(base64.b64encode(header.tobytes() + data_bytes).decode())
+
+        def text_writer_ascii(f):
+            # This write() loop is the bottleneck for the write. Alternatives:
+            # savetxt is super slow:
+            #   np.savetxt(f, data.reshape(-1), fmt=fmt)
+            # joining and writing is a bit faster, but consumes huge amounts of
+            # memory:
+            #   f.write("\n".join(map(fmt.format, data.reshape(-1))))
+            for item in data.reshape(-1):
+                f.write((fmt + "\n").format(item))
+
         if binary:
             da.set("format", "binary")
-            if compression:
-                # compressed write
-                def text_writer(f):
-                    max_block_size = 32768
-                    data_bytes = data.tobytes()
-
-                    # round up
-                    num_blocks = -int(-len(data_bytes) // max_block_size)
-                    last_block_size = (
-                        len(data_bytes) - (num_blocks - 1) * max_block_size
-                    )
-
-                    # It's too bad that we have to keep all blocks in memory. This is
-                    # necessary because the header, written first, needs to know the
-                    # lengths of all blocks. Also, the blocks are encoded _after_ having
-                    # been concatenated.
-                    c = {"lzma": lzma, "zlib": zlib}[compression]
-                    compressed_blocks = [
-                        # This compress is the slowest part of the writer
-                        c.compress(block)
-                        for block in _chunk_it(data_bytes, max_block_size)
-                    ]
-
-                    # collect header
-                    header = np.array(
-                        [num_blocks, max_block_size, last_block_size]
-                        + [len(b) for b in compressed_blocks],
-                        dtype=vtu_to_numpy_type[header_type],
-                    )
-                    f.write(base64.b64encode(header.tobytes()).decode())
-                    f.write(base64.b64encode(b"".join(compressed_blocks)).decode())
-
-            else:
-                # uncompressed write
-                def text_writer(f):
-                    data_bytes = data.tobytes()
-                    # collect header
-                    header = np.array(
-                        len(data_bytes), dtype=vtu_to_numpy_type[header_type]
-                    )
-                    f.write(base64.b64encode(header.tobytes() + data_bytes).decode())
-
+            da.text_writer = (
+                text_writer_compressed if compression else text_writer_uncompressed
+            )
         else:
             da.set("format", "ascii")
-
-            def text_writer(f):
-                # This write() loop is the bottleneck for the write. Alternatives:
-                # savetxt is super slow:
-                #   np.savetxt(f, data.reshape(-1), fmt=fmt)
-                # joining and writing is a bit faster, but consumes huge amounts of
-                # memory:
-                #   f.write("\n".join(map(fmt.format, data.reshape(-1))))
-                for item in data.reshape(-1):
-                    f.write((fmt + "\n").format(item))
-
-        da.text_writer = text_writer
+            da.text_writer = text_writer_ascii
 
     def _polyhedron_face_cells(face_cells):
-        # Define the faces of each cell on the format specfied for VTU Polyhedron cells.
-        # These are defined in Mesh.polyhedron_faces, as block data. The block consists
-        # of a nested list (outer list represents cell, inner is faces for this cells),
-        # where the items of the inner list are the nodes of specific faces.
+        # Define the faces of each cell on the format specified for VTU Polyhedron
+        # cells. These are defined in Mesh.polyhedron_faces, as block data. The block
+        # consists of a nested list (outer list represents cell, inner is faces for this
+        # cells), where the items of the inner list are the nodes of specific faces.
         #
         # The output format is specified at https://vtk.org/Wiki/VTK/Polyhedron_Support
 
@@ -753,7 +774,7 @@ def write(filename, mesh, binary=True, compression="zlib", header_type=None):
 
     grid = ET.SubElement(vtk_file, "UnstructuredGrid")
 
-    total_num_cells = sum([len(c.data) for c in mesh.cells])
+    total_num_cells = sum(len(c.data) for c in mesh.cells)
     piece = ET.SubElement(
         grid,
         "Piece",
@@ -800,12 +821,14 @@ def write(filename, mesh, binary=True, compression="zlib", header_type=None):
 
         else:
             # create connectivity, offset, type arrays
-            connectivity = np.concatenate(
-                [
-                    v.data[:, meshio_to_vtk_order(v.type, v.data.shape[1])].reshape(-1)
-                    for v in mesh.cells
-                ]
-            )
+            connectivity = []
+            for v in mesh.cells:
+                d = v.data
+                new_order = meshio_to_vtk_order(v.type)
+                if new_order is not None:
+                    d = d[:, new_order]
+                connectivity.append(d.flatten())
+            connectivity = np.concatenate(connectivity)
 
             # offset (points to the first element of the next cell)
             offsets = [
@@ -819,12 +842,13 @@ def write(filename, mesh, binary=True, compression="zlib", header_type=None):
 
         # types
         types_array = []
-        for key, v in mesh.cells:
+        for cell_block in mesh.cells:
+            key = cell_block.type
             # some adaptions for polyhedron
             if key.startswith("polyhedron"):
                 # Get face-cell relation on the vtu format. See comments in helper
                 # function for more information of how to specify this.
-                faces_loc, faceoffsets_loc = _polyhedron_face_cells(v)
+                faces_loc, faceoffsets_loc = _polyhedron_face_cells(cell_block.data)
                 # Adjust offsets to global numbering
                 assert faceoffsets is not None
                 if len(faceoffsets) > 0:
@@ -835,7 +859,7 @@ def write(filename, mesh, binary=True, compression="zlib", header_type=None):
                 faceoffsets += faceoffsets_loc
                 key = "polyhedron"
 
-            types_array.append(np.full(len(v), meshio_to_vtk_type[key]))
+            types_array.append(np.full(len(cell_block), meshio_to_vtk_type[key]))
 
         types = np.concatenate(
             types_array
@@ -861,9 +885,16 @@ def write(filename, mesh, binary=True, compression="zlib", header_type=None):
         for name, data in raw_from_cell_data(mesh.cell_data).items():
             numpy_to_xml_array(cd, name, data)
 
+    if mesh.point_sets or mesh.cell_sets:
+        warn(
+            "VTU format cannot write sets. "
+            + "Consider converting them to \\[point,cell]_data.",
+            highlight=False,
+        )
+
     # write_xml(filename, vtk_file, pretty_xml)
     tree = ET.ElementTree(vtk_file)
     tree.write(filename)
 
 
-register("vtu", [".vtu"], read, {"vtu": write})
+register_format("vtu", [".vtu"], read, {"vtu": write})
