@@ -7,10 +7,7 @@ from . import helpers
 
 @pytest.mark.parametrize(
     "mesh",
-    [
-        # helpers.empty_mesh,
-        helpers.tri_mesh
-    ],
+    [helpers.empty_mesh, helpers.tri_mesh],
 )
 @pytest.mark.parametrize(
     "binary, tol",
@@ -20,8 +17,8 @@ from . import helpers
         (True, 1.0e-8),
     ],
 )
-def test_stl(mesh, binary, tol):
+def test_stl(mesh, binary, tol, tmp_path):
     def writer(*args, **kwargs):
         return meshio.stl.write(*args, binary=binary, **kwargs)
 
-    helpers.write_read(writer, meshio.stl.read, mesh, tol)
+    helpers.write_read(tmp_path, writer, meshio.stl.read, mesh, tol)
